/************************************************************************ *********
 * Proiectarea cu Microprocesoare
 * Laboratorul 4
 * Folosirea convertorului analog digital. Caractere custom pe LCD
 *********************************************************************************/

#include <stdio.h>
#include <string.h>
#include <avr\io.h>
#include <avr\interrupt.h>
#define F_CPU 16000000
#include <util\delay.h>


/************************************************************************ 
 * DEFINE-uri pentru parametrizarea modulului LCD
 ************************************************************************/
#define LcdDATA_DDR		DDRB			// Portul pe care conectam firele de date la LCD-ul
#define LcdDATA_PORT	PORTB
#define LcdDATA_PIN		PINB

#define LcdCMD_DDR		DDRB			// Portul pe care conectam firele de comenzi la LCD
#define LcdCMD_PORT		PORTB
#define LcdCMD_PIN		PINB

#define LcdD4			PB0				// Pin-ul pentru firul de date D4 de pe LCD
#define LcdD5			PB1				// Pin-ul pentru firul de date D5 de pe LCD
#define LcdD6			PB2				// Pin-ul pentru firul de date D6 de pe LCD
#define LcdD7			PB3				// Pin-ul pentru firul de date D7 de pe LCD

#define LcdRS			PB4				// Pinul pentru selectare operatie (LCD)
#define LcdRW			PB5				// Pinul pentru Read/ Write (LCD)
#define LcdE			PB6				// Pinul de Enable (LCD)

#define LCD_INSTR_4wire 		0x28 	// 4 fire date, font 5x8
#define LCD_INSTR_display 		0x0C 	// Display On, Cursor On, Blinking On ( 1 Display Cursor Blink )
#define LCD_INSTR_clearDisplay 	0x01 	// Clear Display
#define LCD_INSTR_returnHome 	0x02 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_nextLine 		0xC0 	// Return Cursor and LCD to Home Position
#define LCD_INSTR_gotoCGRAM		0x40	// go to Character Generator RAM

/************************************************************************ 
 * REDIRECTARE stdin, stdout pentru receptie, transmisie seriala
 ************************************************************************/
int uart_putchar(char c, FILE *unused);
int uart_getchar(FILE* f);
FILE mystdout = FDEV_SETUP_STREAM(uart_putchar, NULL, _FDEV_SETUP_WRITE);
FILE mystdin =  FDEV_SETUP_STREAM(NULL, uart_getchar, _FDEV_SETUP_READ);

/************************************************************************ 
 * IMPLEMENTARE FUNCTII TRANSMISIE/RECEPTIE SERIALA, pentru mapare stdin, stdout
 ************************************************************************/
// Functie ce trateaza trimiterea unui caracter pe seriala
int uart_putchar(char c, FILE *unused)      
{
	if (c == '\n')
		uart_putchar('\r', 0);
	loop_until_bit_is_set(UCSRA, UDRE);
	UDR = c;
	return 0;
}

// Functie ce trateaza primirea unui caracter pe seriala
int uart_getchar(FILE* f)
{
	char c;
    
    loop_until_bit_is_set(UCSRA, RXC);
    c = UDR;
    if(c=='\r')
		c='\n';
    uart_putchar(c,NULL); 
    return c;
}

void showTemp(long data) ;

/************************************************************************ 
 * API LCD.. implementare este in partea de jos a fisierului
 ************************************************************************/
void LCD_init();											// Initializare modul LCD.. trebuie apelata inainte de a se face orice operatie cu LCD-ul
void LCD_writeInstruction(unsigned char _instruction);		// Trimite o instructiune catre lcd (vezi datasheet)
void LCD_writeData(unsigned char _data);					// Trimite date catre LCD pentru afisare
void LCD_write(unsigned char _byte);						// trimite un bute catre LCD la modul general (nu conteaza daca e instructiune sau date)
void LCD_waitNotBusy();										// Asteptam pana cand lcd-ul devine disponibil pt o noua comanda
void LCD_waitInstructions(unsigned char _instructions);		// Asteapta un numar de cicli de tact.. loop

// tine definitiile caracterelor pentru progress bar
char progress_bar[6 * 8] =
{
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 ,	// gol
	0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10, 0x10 ,	// o bara verticala
	0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18 ,	// doua
	0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C, 0x1C ,	// trei
	0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E, 0x1E ,	// patru
	0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F, 0x1F		// cinci
};
char mes[]="Salut sunt niste text care vreau sa apar aici";
long data1,  data2, temp, life, avg, fulavg, maxt, mint, sett;

// initializarea convertorului analog digital
void ADC_init1()
{
	// AVCC cu condesator pe AREF; PA0 ca input single ended; rezultat left-adjusted (8 biti de date)
	ADMUX = (1<<REFS0) | (1<<ADLAR) ;
	// activeaza ADC si intreruperea de conversie. Prescaler de ceas la maximum
	ADCSRA = (1<<ADEN) | (1<<ADIE) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0);
}

// initializarea convertorului analog digital
void ADC_init2()
{
	// AVCC cu condesator pe AREF; PA1 ca input single ended; rezultat left-adjusted (8 biti de date)
	ADMUX = (1<<REFS0) | (1<<ADLAR) | (1<<MUX0);
	// activeaza ADC si intreruperea de conversie. Prescaler de ceas la maximum
	ADCSRA = (1<<ADEN) | (1<<ADIE) | (1<<ADPS2) | (1<<ADPS1) | (1<<ADPS0);
}
 
 // Exemplu de functie ce se apeleaza la interval de 1 secunda, intrerupere generata de timerul 1 (vezi laboratorul 2) 
SIGNAL(SIG_OUTPUT_COMPARE1A)
{	
	long data=0;
//adaptez variabilele
	life++;
	sett++;

	avg=0;
		//sterg ecranul
		LCD_writeInstruction(LCD_INSTR_clearDisplay);	
		
		
		
		
		
		data=data1;
		//scriu la lcd rezultatul obtinut
		LCD_writeData('T');
		LCD_writeData('1');
		LCD_writeData(':');
		showTemp(data);
		//updatez variabile
		
		if (data>0 && mint>data) mint=data;
		if (maxt<data) maxt=data;
		//avg=data;
		
		////dezactivam ADC
		//ADCSRA =0;
		//a doua temperatura
		
		
		
		//scriu la lcd rezultatul obtinut
		//ADCSRA=0;
		data=data2;
		LCD_writeData(' ');
		LCD_writeData('T');
		LCD_writeData('2');
		LCD_writeData(':');
		showTemp(data);
		//updatez variabile
		if (data>0 && mint>data) mint=data;
		if (maxt<data) maxt=data;
		avg=avg+data;
		//avg=avg/2;
		//printez la LCD temperatura medie
		// trecem pe linia urmatoare
		LCD_writeInstruction(LCD_INSTR_nextLine);
		LCD_writeData('M');
		LCD_writeData('e');
		LCD_writeData('d');
		LCD_writeData(':');
		showTemp(avg);
		//si setata
		LCD_writeData(' ');
		LCD_writeData('S');
		LCD_writeData(':');
		showTemp(sett);
}
 
//functie care reprezinta temperatura pe LCD
void showTemp(long t) {
	// afisez cifra zecilor
	t=t%1000;
	LCD_writeData('0' + t/100);
	t %= 100;
	// afisez cifra unitatilor
	LCD_writeData('0' + t/10);
	t %= 10;
	// afisez virgula
	LCD_writeData('.');
	// afisez prima zecimala
	LCD_writeData('0' + t);
	
	
	}
 
// intreruperea de conversie
SIGNAL(SIG_ADC)
{
	int  i,j;
	if (temp==0){
		data1=(ADCL>>6)+(ADCH<<2);
		data1=(long)data1*5000/1024;
		ADCSRA=0;
		_delay_ms(50);		
		
		ADC_init2();
		
		}
	else {
		data2=(ADCL>>6)+(ADCH<<2);
		data2=(long)data2*5000/1024;
		ADCSRA=0;
		_delay_ms(50);		
		
		//prima temperatura
		ADC_init1();
	}
		
	temp=1-temp;
		_delay_ms(500);
	// cer ADC-ului sa faca o conversie
		ADCSRA |= (1<<ADSC);
}

int main()
{
	//initializari
	int i;
	
	char welcome[]="Starting up";
	char message[33];		
	double tmp;
	life=0; avg=0; fulavg=0;
	maxt=0; mint=123456;
	sett=200;
	temp=0;
	data1=0; data2=0;
	// Initializare modul USART pt comunicatie la 9600bps, 8 biti de date, la 16MHz											
	UBRRH = 0;				
	UBRRL = 103;
	UCSRB = 0x18;
	UCSRC = _BV(URSEL) | _BV(UCSZ1) | _BV(UCSZ0);
	stdin = &mystdin;
	stdout = &mystdout;

	// Initializare timer 1 compare match A pentru a avea un timer la nivel de 1 secunda, 16MHz
	OCR1A = 15624;
	TCCR1A = 0;
	TCCR1B = (1<<WGM12)|(1<<CS12)|(1<<CS10);
	TIMSK |= _BV(OCIE1A);
	
	
	LCD_init();
	//sterg ecranul
	LCD_writeInstruction(LCD_INSTR_clearDisplay);	
	for(i=0; i<11; i++) LCD_writeData(welcome[i]);
	_delay_ms(500);

	//initializare intreruperi
	sei();
	ADC_init1();
	// cer ADC-ului sa faca o conversie
		ADCSRA |= (1<<ADSC);
	for(;;)
	{
		printf("Enter 1 for setting maximum temperature or 2 for status information\n");
		scanf("%d",&i);
		if (i==1) {
			printf("Enter the maximum allowed temperature\n");
			scanf("%lf",&tmp);
			printf("Temperature set to %lf\n",tmp);
			sett=(long)(tmp*10);
			}
		else if (i==2) {
			printf("System has been running for %ld hours, %ld minutes %ld seconds\n",life/3600, (life%3600)/60, 
				(life%3600)%60);
			printf("Minimum and maximum recorded temperatures are: %ld %ld\n", mint, maxt);
			printf("Maximum allowed temperature set to %ld\n",sett);
			}
					
	}
	return 0;
}

/************************************************************************ 
 * IMPLEMENTARE API LCD.. implementare este in partea de jos a fisierului
 ************************************************************************/

void LCD_init()
{
	LcdDATA_DDR |=  (1<<LcdD4)|(1<<LcdD5)|(1<<LcdD6)|(1<<LcdD7);	// setam pinii de date ca pini de iesire
	LcdCMD_DDR  |=  (1<<LcdRS)|(1<<LcdRW)|(1<<LcdE);				// setam pinii de comenzi ca pini de iesire

	LCD_waitNotBusy();
	
	LcdCMD_PORT   &= ~(1<<LcdRS);								// Setam linia RS pe low
	LcdCMD_PORT   &= ~(1<<LcdRW);								// Setam linia RW pe low (acum suntem in modul de trimis instructiuni)
	LcdDATA_PORT  &= ~(1<<LcdD4)&~(1<<LcdD6)&~(1<<LcdD7); 		// Specificam ca vrem 4 fire de date, prima comanda (LcdD5 activ, restul nu)
	LcdDATA_PORT  |=  (1<<LcdD5);								// setam pinii de comenzi ca pini de iesire
	
	LcdCMD_PORT |=  (1<<LcdE);					// Setam linia E(nable) pe high; aceasta ii specifica LCD-ului sa preia datele
	LCD_waitInstructions(6);					// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE );					// Setam linia E(nable) pe low; transferul s-a terminat

	LCD_writeInstruction(LCD_INSTR_4wire); 		// Incarcam comanda: 4 bit data, 2 lines, 5x8 font
	LCD_writeInstruction(LCD_INSTR_display); 	// Display On, Cursor On, Blinking On
	LCD_writeInstruction(0x06);					// Increment, no shift
	LCD_writeInstruction(0x01);					// Clear Display
}

void LCD_writeInstruction(unsigned char _instruction)
{
	LCD_waitNotBusy();						// asteptam ca LCD-ul sa fie liber sa primeasca comenzile	
	LcdCMD_PORT &= ~(1 << LcdRS);			// setam pinul RS pe low.. low=instructiuni, high=date
	LcdCMD_PORT &= ~(1 << LcdRW);			// setam pinul RW pe low (suntem in modul de comenzi acum)
	LCD_write(_instruction);				// apelam procedura ce trimite byte-ul pe firele de date
}


void LCD_writeData(unsigned char _data)
{
	LCD_waitNotBusy();						// asteptam ca LCD-ul sa fie liber sa primeasca comenzile	
	LcdCMD_PORT |= (1 << LcdRS);			// setam pinul RS pe high
	LcdCMD_PORT &= ~(1 << LcdRW);			// setam pinul RW pe low (suntem in modul de comenzi acum)
	LCD_write(_data);						// apelam procedura ce trimite byte-ul pe firele de date
}

void LCD_write(unsigned char _byte)
{
	unsigned char _byte2;
	
	_byte2 = _byte>>4;

	LcdDATA_PORT &= ~(1<<LcdD4);
	if ( bit_is_set( _byte2, 0 ) )
		LcdDATA_PORT |= (1<<LcdD4);
		
	LcdDATA_PORT &= ~(1<<LcdD5);
	if ( bit_is_set( _byte2, 1 ) )
		LcdDATA_PORT |= (1<<LcdD5);

	LcdDATA_PORT &= ~(1<<LcdD6);
	if ( bit_is_set( _byte2, 2 ) )
		LcdDATA_PORT |= (1<<LcdD6);

	LcdDATA_PORT &= ~(1<<LcdD7);
	if ( bit_is_set( _byte2, 3 ) )
		LcdDATA_PORT |= (1<<LcdD7);

	LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
	LCD_waitInstructions(6);						// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low

	LCD_waitInstructions(6);						// Asteptam o perioada de timp T
	
	LcdDATA_PORT &= ~(1<<LcdD4);
	if ( bit_is_set( _byte, 0 ) )
		LcdDATA_PORT |= (1<<LcdD4);
	
	LcdDATA_PORT &= ~(1<<LcdD5);
	if ( bit_is_set( _byte, 1 ) )
		LcdDATA_PORT |= (1<<LcdD5);

	LcdDATA_PORT &= ~(1<<LcdD6);
	if ( bit_is_set( _byte, 2 ) )
		LcdDATA_PORT |= (1<<LcdD6);

	LcdDATA_PORT &= ~(1<<LcdD7);
	if ( bit_is_set( _byte, 3 ) )
		LcdDATA_PORT |= (1<<LcdD7);

	LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
	LCD_waitInstructions(6);						// Asteptam o perioada de timp T
	LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low
}

void LCD_waitNotBusy()
{
	unsigned char _loop = 1;

	/*while (_loop)
	{
		LcdDATA_DDR &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7);	// Setam pinii de date de la LCD pe in pt a citi busy flag
		LcdDATA_PORT &= ~(1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); 	// Dezactivam pullup resistor pentru pinii de in
		
		LcdCMD_PORT &= ~(1<<LcdE);					// Setam pin-ul e pe low; ar trebui sa fie deja low, doar ne asiguram
		LcdCMD_PORT &= ~(1<<LcdRS);					// Setam pinul RS pe low
		LcdCMD_PORT |=  (1<<LcdRW);					// Setam pinul RW pe high (acum suntem in modul de interogare busy/adr)
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		_loop = LcdDATA_PIN & (1<<LcdD7);				// Citim busy flag-ul
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low
		
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		
		LcdCMD_PORT |= (1<<LcdE);						// Setam Pinul E pe high
		LCD_waitInstructions(6);						// Asteptam o perioada de timp T
		LcdCMD_PORT &= ~(1<<LcdE);						// Setam Pinul E pe low

		LcdDATA_DDR |= (1<<LcdD4 | 1<<LcdD5 | 1<<LcdD6 | 1<<LcdD7); // Setam Portul de LCD ca port de iesire la loc
	}
	//*/
	_delay_ms(5);
}

void LCD_waitInstructions(unsigned char _instructions)
{
	while (_instructions--)
		;
}
